import uuid
import threading
import os
from asyncio import Future
from typing import Dict, Generic, Iterable, Union
from typing_extensions import TypeVarTuple, Unpack
from hedra.core.engines.types.common.types import RequestTypes
from .store import ActionsStore
from .config import Config
from .client_types import  (
    HTTPClient,
    HTTP2Client,
    GRPCClient,
    GraphQLClient,
    GraphQLHTTP2Client,
    WebsocketClient,
    PlaywrightClient,
    TaskClient,
    UDPClient
)

from .plugins_store import PluginsStore


T = TypeVarTuple('T')


class Client(Generic[Unpack[T]]):

    def __init__(
        self,
        graph_name: str,
        graph_id: str,
        stage_name: str,
        stage_id: str
    ) -> None:

        self.client_id = str(uuid.uuid4())
        self.graph_name = graph_name
        self.graph_id = graph_id
        self.stage_name = stage_name
        self.stage_id = stage_id

        self.next_name = None
        self.intercept = False

        self._config: Config = None
        self._http = HTTPClient
        self._http2 = HTTP2Client
        self._grpc = GRPCClient
        self._graphql = GraphQLClient
        self._graphqlh2 = GraphQLHTTP2Client
        self._websocket = WebsocketClient
        self._playwright = PlaywrightClient
        self._udp = UDPClient
        self._task = TaskClient

        self.clients = {}
        self._plugin = PluginsStore[Unpack[T]](self.metadata_string)

        self.actions = ActionsStore(self.metadata_string)

    def __getitem__(self, key: str):
        return self.clients.get(key)

    def __setitem__(self, key, value):
        self.clients[key] = value

    def get_waiters(self) -> Iterable[Future]:
        for session in self.clients.values():
            if session.waiter:
                yield session.waiter

    @property
    def thread_id(self) -> int:
        return threading.current_thread().ident

    @property
    def process_id(self) -> int:
        return os.getpid()

    @property
    def metadata_string(self):
        return f'Graph - {self.graph_name}:{self.graph_id} - thread:{self.thread_id} - process:{self.process_id} - Stage: {self.stage_name}:{self.stage_id} - '

    @property
    def plugin(self) -> Dict[str, Union[Unpack[T]]]:
        self._plugin._config = self._config
        self._plugin.actions = self.actions
        self._plugin.metadata_string = self.metadata_string
        self._plugin.actions.waiter = self.actions.waiter
        self._plugin.actions.current_stage = self.actions.current_stage
        self._plugin.intercept = self.intercept
        self._plugin.next_name = self.next_name

        return self._plugin

    @property
    def http(self):
        if self._http.initialized is False:
            self._http = self._http(self._config)
            self._http.metadata_string = self.metadata_string
            self._http.actions = self.actions
            self.clients[RequestTypes.HTTP] = self._http

        self._http.next_name = self.next_name
        self._http.intercept = self.intercept
        return self._http

    @property
    def http2(self):
        if self._http2.initialized is False:
            self._http2 = self._http2(self._config)
            self._http2.metadata_string = self.metadata_string
            self._http2.actions = self.actions
            self.clients[RequestTypes.HTTP2] = self._http2
       
        self._http2.next_name = self.next_name
        self._http2.intercept = self.intercept
        return self._http2

    @property
    def grpc(self):
        if self._grpc.initialized is False:
            self._grpc = self._grpc(self._config)
            self._grpc.metadata_string = self.metadata_string
            self._grpc.actions = self.actions
            self.clients[RequestTypes.GRPC] = self._grpc

        self._grpc.next_name = self.next_name
        self._grpc.intercept = self.intercept
        return self._grpc

    @property
    def graphql(self):
        if self._graphql.initialized is False:
            self._graphql = self._graphql(self._config)
            self._graphql.metadata_string = self.metadata_string
            self._graphql.actions = self.actions
            self.clients[RequestTypes.GRAPHQL] = self._graphql

        self._graphql.next_name = self.next_name
        self._graphql.intercept = self.intercept
        return self._graphql

    @property
    def graphqlh2(self):
        if self._graphqlh2.initialized is False:
            self._graphqlh2 = self._graphqlh2(self._config)
            self._graphqlh2.metadata_string = self.metadata_string
            self._graphqlh2.actions = self.actions
            self.clients[RequestTypes.GRAPHQL_HTTP2] = self._graphqlh2

        self._graphqlh2.next_name = self.next_name
        self._graphqlh2.intercept = self.intercept
        return self._graphql

    @property
    def websocket(self):
        if self._websocket.initialized is False:
            self._websocket = self._websocket(self._config)
            self._websocket.metadata_string = self.metadata_string
            self._websocket.actions = self.actions
            self.clients[RequestTypes.WEBSOCKET] = self._websocket

        self._websocket.next_name = self.next_name
        self._websocket.intercept = self.intercept
        return self._websocket

    @property
    def playwright(self):
        if self._playwright.initialized is False:
            self._playwright = self._playwright(self._config)
            self._playwright.metadata_string = self.metadata_string
            self._playwright.actions = self.actions
            self.clients[RequestTypes.PLAYWRIGHT] = self._playwright

        self._playwright.next_name = self.next_name
        self._playwright.intercept = self.intercept
        return self._playwright

    @property
    def udp(self):
        if self._udp.initialized is False:
            self._udp = self._udp(self._config)
            self._udp.metadata_string = self.metadata_string
            self._udp.actions = self.actions
            self.clients[RequestTypes.UDP] = self._udp

        self._udp.next_name = self.next_name
        self._udp.intercept = self.intercept
        return self._udp
    
    @property
    def task(self):
        if self._task.initialized is False:
            self._task = self._task(self._config)
            self._task.metadata_string = self.metadata_string
            self.clients[RequestTypes.TASK] = self._task

        self._task.next_name = self.next_name
        return self._task


        


        
        



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: leader.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='leader.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x0cleader.proto\x1a\x1cgoogle/protobuf/struct.proto\"V\n\x15WorkerRegisterRequest\x12\x14\n\x0chost_address\x18\x01 \x01(\t\x12\x0f\n\x07host_id\x18\x02 \x01(\t\x12\x16\n\x0eserver_address\x18\x03 \x01(\t\"l\n\x16WorkerRegisterResponse\x12\x14\n\x0chost_address\x18\x01 \x01(\t\x12\x0f\n\x07host_id\x18\x02 \x01(\t\x12+\n\njob_config\x18\x03 \x01(\x0b\x32\x17.google.protobuf.Struct\"<\n\x13WorkerUpdateRequest\x12\x14\n\x0chost_address\x18\x01 \x01(\t\x12\x0f\n\x07host_id\x18\x02 \x01(\x03\"i\n\x14WorkerUpdateResponse\x12\x14\n\x0chost_address\x18\x01 \x01(\t\x12\x0f\n\x07host_id\x18\x02 \x01(\t\x12*\n\tcompleted\x18\x03 \x01(\x0b\x32\x17.google.protobuf.Struct\"Y\n\x17WorkerDeregisterRequest\x12\x14\n\x0chost_address\x18\x01 \x01(\t\x12\x0f\n\x07host_id\x18\x02 \x01(\t\x12\x17\n\x0freporter_fields\x18\x03 \x03(\t\"A\n\x18WorkerDeregisterResponse\x12\x14\n\x0chost_address\x18\x01 \x01(\t\x12\x0f\n\x07host_id\x18\x02 \x01(\t\">\n\x15LeaderRegisterRequest\x12\x14\n\x0chost_address\x18\x01 \x01(\t\x12\x0f\n\x07host_id\x18\x02 \x01(\t\"?\n\x16LeaderRegisterResponse\x12\x14\n\x0chost_address\x18\x01 \x01(\t\x12\x0f\n\x07host_id\x18\x02 \x01(\t\":\n\x11PollLeaderRequest\x12\x14\n\x0chost_address\x18\x01 \x01(\t\x12\x0f\n\x07host_id\x18\x02 \x01(\t\"N\n\x12PollLeaderResponse\x12\x14\n\x0chost_address\x18\x01 \x01(\t\x12\x0f\n\x07host_id\x18\x02 \x01(\t\x12\x11\n\tcompleted\x18\x03 \x01(\x08\"d\n\x14PollLeaderJobRequest\x12\x14\n\x0chost_address\x18\x01 \x01(\t\x12\x0f\n\x07host_id\x18\x02 \x01(\t\x12\x0e\n\x06job_id\x18\x03 \x01(\t\x12\x15\n\rcurrent_stage\x18\x04 \x01(\t\"Q\n\x15PollLeaderJobResponse\x12\x14\n\x0chost_address\x18\x01 \x01(\t\x12\x0f\n\x07host_id\x18\x02 \x01(\t\x12\x11\n\tcompleted\x18\x03 \x01(\x08\"f\n\x14PipelineStageRequest\x12\x14\n\x0chost_address\x18\x01 \x01(\t\x12\x0f\n\x07host_id\x18\x02 \x01(\t\x12\x0e\n\x06job_id\x18\x03 \x01(\t\x12\x17\n\x0fstage_completed\x18\x04 \x01(\t\">\n\x15PipelineStageResponse\x12\x14\n\x0chost_address\x18\x01 \x01(\t\x12\x0f\n\x07host_id\x18\x02 \x01(\t\"~\n\x13NewJobLeaderRequest\x12\x14\n\x0chost_address\x18\x01 \x01(\t\x12\x0f\n\x07host_id\x18\x02 \x01(\t\x12+\n\njob_config\x18\x03 \x01(\x0b\x32\x17.google.protobuf.Struct\x12\x13\n\x0bjob_timeout\x18\x04 \x01(\x03\"M\n\x14NewJobLeaderResponse\x12\x14\n\x0chost_address\x18\x01 \x01(\t\x12\x0f\n\x07host_id\x18\x02 \x01(\t\x12\x0e\n\x06job_id\x18\x03 \x01(\t\"O\n\x16LeaderHeartbeatRequest\x12\x14\n\x0chost_address\x18\x01 \x01(\t\x12\x0f\n\x07host_id\x18\x02 \x01(\t\x12\x0e\n\x06status\x18\x03 \x01(\t\"P\n\x17LeaderHeartbeatResponse\x12\x14\n\x0chost_address\x18\x01 \x01(\t\x12\x0f\n\x07host_id\x18\x02 \x01(\t\x12\x0e\n\x06status\x18\x03 \x01(\t\"\xa8\x01\n\x10LeaderJobRequest\x12\x14\n\x0chost_address\x18\x01 \x01(\t\x12\x0f\n\x07host_id\x18\x02 \x01(\t\x12\x0e\n\x06job_id\x18\x03 \x01(\t\x12+\n\njob_config\x18\x04 \x01(\x0b\x32\x17.google.protobuf.Struct\x12\x30\n\x0fpipeline_status\x18\x05 \x01(\x0b\x32\x17.google.protobuf.Struct\"\xa9\x01\n\x11LeaderJobResponse\x12\x14\n\x0chost_address\x18\x01 \x01(\t\x12\x0f\n\x07host_id\x18\x02 \x01(\t\x12\x0e\n\x06job_id\x18\x03 \x01(\t\x12+\n\njob_config\x18\x04 \x01(\x0b\x32\x17.google.protobuf.Struct\x12\x30\n\x0fpipeline_status\x18\x05 \x01(\x0b\x32\x17.google.protobuf.Struct2\xa5\x05\n\x11\x44istributedServer\x12>\n\tAddClient\x12\x16.WorkerRegisterRequest\x1a\x17.WorkerRegisterResponse\"\x00\x12:\n\tGetUpdate\x12\x14.WorkerUpdateRequest\x1a\x15.WorkerUpdateResponse\"\x00\x12\x45\n\x0cRemoveClient\x12\x18.WorkerDeregisterRequest\x1a\x19.WorkerDeregisterResponse\"\x00\x12\x43\n\x0eRegisterLeader\x12\x16.LeaderRegisterRequest\x1a\x17.LeaderRegisterResponse\"\x00\x12\x37\n\nPollLeader\x12\x12.PollLeaderRequest\x1a\x13.PollLeaderResponse\"\x00\x12\x45\n\x12MarkStageCompleted\x12\x15.PipelineStageRequest\x1a\x16.PipelineStageResponse\"\x00\x12@\n\rPollLeaderJob\x12\x15.PollLeaderJobRequest\x1a\x16.PollLeaderJobResponse\"\x00\x12=\n\x0c\x43reateNewJob\x12\x14.NewJobLeaderRequest\x1a\x15.NewJobLeaderResponse\"\x00\x12K\n\x14\x43heckLeaderHeartbeat\x12\x17.LeaderHeartbeatRequest\x1a\x18.LeaderHeartbeatResponse\"\x00\x12:\n\x0fShareCurrentJob\x12\x11.LeaderJobRequest\x1a\x12.LeaderJobResponse\"\x00\x62\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_struct__pb2.DESCRIPTOR,])




_WORKERREGISTERREQUEST = _descriptor.Descriptor(
  name='WorkerRegisterRequest',
  full_name='WorkerRegisterRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='host_address', full_name='WorkerRegisterRequest.host_address', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='host_id', full_name='WorkerRegisterRequest.host_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='server_address', full_name='WorkerRegisterRequest.server_address', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=46,
  serialized_end=132,
)


_WORKERREGISTERRESPONSE = _descriptor.Descriptor(
  name='WorkerRegisterResponse',
  full_name='WorkerRegisterResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='host_address', full_name='WorkerRegisterResponse.host_address', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='host_id', full_name='WorkerRegisterResponse.host_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='job_config', full_name='WorkerRegisterResponse.job_config', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=134,
  serialized_end=242,
)


_WORKERUPDATEREQUEST = _descriptor.Descriptor(
  name='WorkerUpdateRequest',
  full_name='WorkerUpdateRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='host_address', full_name='WorkerUpdateRequest.host_address', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='host_id', full_name='WorkerUpdateRequest.host_id', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=244,
  serialized_end=304,
)


_WORKERUPDATERESPONSE = _descriptor.Descriptor(
  name='WorkerUpdateResponse',
  full_name='WorkerUpdateResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='host_address', full_name='WorkerUpdateResponse.host_address', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='host_id', full_name='WorkerUpdateResponse.host_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='completed', full_name='WorkerUpdateResponse.completed', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=306,
  serialized_end=411,
)


_WORKERDEREGISTERREQUEST = _descriptor.Descriptor(
  name='WorkerDeregisterRequest',
  full_name='WorkerDeregisterRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='host_address', full_name='WorkerDeregisterRequest.host_address', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='host_id', full_name='WorkerDeregisterRequest.host_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='reporter_fields', full_name='WorkerDeregisterRequest.reporter_fields', index=2,
      number=3, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=413,
  serialized_end=502,
)


_WORKERDEREGISTERRESPONSE = _descriptor.Descriptor(
  name='WorkerDeregisterResponse',
  full_name='WorkerDeregisterResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='host_address', full_name='WorkerDeregisterResponse.host_address', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='host_id', full_name='WorkerDeregisterResponse.host_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=504,
  serialized_end=569,
)


_LEADERREGISTERREQUEST = _descriptor.Descriptor(
  name='LeaderRegisterRequest',
  full_name='LeaderRegisterRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='host_address', full_name='LeaderRegisterRequest.host_address', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='host_id', full_name='LeaderRegisterRequest.host_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=571,
  serialized_end=633,
)


_LEADERREGISTERRESPONSE = _descriptor.Descriptor(
  name='LeaderRegisterResponse',
  full_name='LeaderRegisterResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='host_address', full_name='LeaderRegisterResponse.host_address', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='host_id', full_name='LeaderRegisterResponse.host_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=635,
  serialized_end=698,
)


_POLLLEADERREQUEST = _descriptor.Descriptor(
  name='PollLeaderRequest',
  full_name='PollLeaderRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='host_address', full_name='PollLeaderRequest.host_address', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='host_id', full_name='PollLeaderRequest.host_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=700,
  serialized_end=758,
)


_POLLLEADERRESPONSE = _descriptor.Descriptor(
  name='PollLeaderResponse',
  full_name='PollLeaderResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='host_address', full_name='PollLeaderResponse.host_address', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='host_id', full_name='PollLeaderResponse.host_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='completed', full_name='PollLeaderResponse.completed', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=760,
  serialized_end=838,
)


_POLLLEADERJOBREQUEST = _descriptor.Descriptor(
  name='PollLeaderJobRequest',
  full_name='PollLeaderJobRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='host_address', full_name='PollLeaderJobRequest.host_address', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='host_id', full_name='PollLeaderJobRequest.host_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='job_id', full_name='PollLeaderJobRequest.job_id', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='current_stage', full_name='PollLeaderJobRequest.current_stage', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=840,
  serialized_end=940,
)


_POLLLEADERJOBRESPONSE = _descriptor.Descriptor(
  name='PollLeaderJobResponse',
  full_name='PollLeaderJobResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='host_address', full_name='PollLeaderJobResponse.host_address', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='host_id', full_name='PollLeaderJobResponse.host_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='completed', full_name='PollLeaderJobResponse.completed', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=942,
  serialized_end=1023,
)


_PIPELINESTAGEREQUEST = _descriptor.Descriptor(
  name='PipelineStageRequest',
  full_name='PipelineStageRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='host_address', full_name='PipelineStageRequest.host_address', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='host_id', full_name='PipelineStageRequest.host_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='job_id', full_name='PipelineStageRequest.job_id', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='stage_completed', full_name='PipelineStageRequest.stage_completed', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1025,
  serialized_end=1127,
)


_PIPELINESTAGERESPONSE = _descriptor.Descriptor(
  name='PipelineStageResponse',
  full_name='PipelineStageResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='host_address', full_name='PipelineStageResponse.host_address', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='host_id', full_name='PipelineStageResponse.host_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1129,
  serialized_end=1191,
)


_NEWJOBLEADERREQUEST = _descriptor.Descriptor(
  name='NewJobLeaderRequest',
  full_name='NewJobLeaderRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='host_address', full_name='NewJobLeaderRequest.host_address', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='host_id', full_name='NewJobLeaderRequest.host_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='job_config', full_name='NewJobLeaderRequest.job_config', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='job_timeout', full_name='NewJobLeaderRequest.job_timeout', index=3,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1193,
  serialized_end=1319,
)


_NEWJOBLEADERRESPONSE = _descriptor.Descriptor(
  name='NewJobLeaderResponse',
  full_name='NewJobLeaderResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='host_address', full_name='NewJobLeaderResponse.host_address', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='host_id', full_name='NewJobLeaderResponse.host_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='job_id', full_name='NewJobLeaderResponse.job_id', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1321,
  serialized_end=1398,
)


_LEADERHEARTBEATREQUEST = _descriptor.Descriptor(
  name='LeaderHeartbeatRequest',
  full_name='LeaderHeartbeatRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='host_address', full_name='LeaderHeartbeatRequest.host_address', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='host_id', full_name='LeaderHeartbeatRequest.host_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='status', full_name='LeaderHeartbeatRequest.status', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1400,
  serialized_end=1479,
)


_LEADERHEARTBEATRESPONSE = _descriptor.Descriptor(
  name='LeaderHeartbeatResponse',
  full_name='LeaderHeartbeatResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='host_address', full_name='LeaderHeartbeatResponse.host_address', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='host_id', full_name='LeaderHeartbeatResponse.host_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='status', full_name='LeaderHeartbeatResponse.status', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1481,
  serialized_end=1561,
)


_LEADERJOBREQUEST = _descriptor.Descriptor(
  name='LeaderJobRequest',
  full_name='LeaderJobRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='host_address', full_name='LeaderJobRequest.host_address', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='host_id', full_name='LeaderJobRequest.host_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='job_id', full_name='LeaderJobRequest.job_id', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='job_config', full_name='LeaderJobRequest.job_config', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='pipeline_status', full_name='LeaderJobRequest.pipeline_status', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1564,
  serialized_end=1732,
)


_LEADERJOBRESPONSE = _descriptor.Descriptor(
  name='LeaderJobResponse',
  full_name='LeaderJobResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='host_address', full_name='LeaderJobResponse.host_address', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='host_id', full_name='LeaderJobResponse.host_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='job_id', full_name='LeaderJobResponse.job_id', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='job_config', full_name='LeaderJobResponse.job_config', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='pipeline_status', full_name='LeaderJobResponse.pipeline_status', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1735,
  serialized_end=1904,
)

_WORKERREGISTERRESPONSE.fields_by_name['job_config'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_WORKERUPDATERESPONSE.fields_by_name['completed'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_NEWJOBLEADERREQUEST.fields_by_name['job_config'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_LEADERJOBREQUEST.fields_by_name['job_config'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_LEADERJOBREQUEST.fields_by_name['pipeline_status'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_LEADERJOBRESPONSE.fields_by_name['job_config'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_LEADERJOBRESPONSE.fields_by_name['pipeline_status'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
DESCRIPTOR.message_types_by_name['WorkerRegisterRequest'] = _WORKERREGISTERREQUEST
DESCRIPTOR.message_types_by_name['WorkerRegisterResponse'] = _WORKERREGISTERRESPONSE
DESCRIPTOR.message_types_by_name['WorkerUpdateRequest'] = _WORKERUPDATEREQUEST
DESCRIPTOR.message_types_by_name['WorkerUpdateResponse'] = _WORKERUPDATERESPONSE
DESCRIPTOR.message_types_by_name['WorkerDeregisterRequest'] = _WORKERDEREGISTERREQUEST
DESCRIPTOR.message_types_by_name['WorkerDeregisterResponse'] = _WORKERDEREGISTERRESPONSE
DESCRIPTOR.message_types_by_name['LeaderRegisterRequest'] = _LEADERREGISTERREQUEST
DESCRIPTOR.message_types_by_name['LeaderRegisterResponse'] = _LEADERREGISTERRESPONSE
DESCRIPTOR.message_types_by_name['PollLeaderRequest'] = _POLLLEADERREQUEST
DESCRIPTOR.message_types_by_name['PollLeaderResponse'] = _POLLLEADERRESPONSE
DESCRIPTOR.message_types_by_name['PollLeaderJobRequest'] = _POLLLEADERJOBREQUEST
DESCRIPTOR.message_types_by_name['PollLeaderJobResponse'] = _POLLLEADERJOBRESPONSE
DESCRIPTOR.message_types_by_name['PipelineStageRequest'] = _PIPELINESTAGEREQUEST
DESCRIPTOR.message_types_by_name['PipelineStageResponse'] = _PIPELINESTAGERESPONSE
DESCRIPTOR.message_types_by_name['NewJobLeaderRequest'] = _NEWJOBLEADERREQUEST
DESCRIPTOR.message_types_by_name['NewJobLeaderResponse'] = _NEWJOBLEADERRESPONSE
DESCRIPTOR.message_types_by_name['LeaderHeartbeatRequest'] = _LEADERHEARTBEATREQUEST
DESCRIPTOR.message_types_by_name['LeaderHeartbeatResponse'] = _LEADERHEARTBEATRESPONSE
DESCRIPTOR.message_types_by_name['LeaderJobRequest'] = _LEADERJOBREQUEST
DESCRIPTOR.message_types_by_name['LeaderJobResponse'] = _LEADERJOBRESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

WorkerRegisterRequest = _reflection.GeneratedProtocolMessageType('WorkerRegisterRequest', (_message.Message,), {
  'DESCRIPTOR' : _WORKERREGISTERREQUEST,
  '__module__' : 'leader_pb2'
  # @@protoc_insertion_point(class_scope:WorkerRegisterRequest)
  })
_sym_db.RegisterMessage(WorkerRegisterRequest)

WorkerRegisterResponse = _reflection.GeneratedProtocolMessageType('WorkerRegisterResponse', (_message.Message,), {
  'DESCRIPTOR' : _WORKERREGISTERRESPONSE,
  '__module__' : 'leader_pb2'
  # @@protoc_insertion_point(class_scope:WorkerRegisterResponse)
  })
_sym_db.RegisterMessage(WorkerRegisterResponse)

WorkerUpdateRequest = _reflection.GeneratedProtocolMessageType('WorkerUpdateRequest', (_message.Message,), {
  'DESCRIPTOR' : _WORKERUPDATEREQUEST,
  '__module__' : 'leader_pb2'
  # @@protoc_insertion_point(class_scope:WorkerUpdateRequest)
  })
_sym_db.RegisterMessage(WorkerUpdateRequest)

WorkerUpdateResponse = _reflection.GeneratedProtocolMessageType('WorkerUpdateResponse', (_message.Message,), {
  'DESCRIPTOR' : _WORKERUPDATERESPONSE,
  '__module__' : 'leader_pb2'
  # @@protoc_insertion_point(class_scope:WorkerUpdateResponse)
  })
_sym_db.RegisterMessage(WorkerUpdateResponse)

WorkerDeregisterRequest = _reflection.GeneratedProtocolMessageType('WorkerDeregisterRequest', (_message.Message,), {
  'DESCRIPTOR' : _WORKERDEREGISTERREQUEST,
  '__module__' : 'leader_pb2'
  # @@protoc_insertion_point(class_scope:WorkerDeregisterRequest)
  })
_sym_db.RegisterMessage(WorkerDeregisterRequest)

WorkerDeregisterResponse = _reflection.GeneratedProtocolMessageType('WorkerDeregisterResponse', (_message.Message,), {
  'DESCRIPTOR' : _WORKERDEREGISTERRESPONSE,
  '__module__' : 'leader_pb2'
  # @@protoc_insertion_point(class_scope:WorkerDeregisterResponse)
  })
_sym_db.RegisterMessage(WorkerDeregisterResponse)

LeaderRegisterRequest = _reflection.GeneratedProtocolMessageType('LeaderRegisterRequest', (_message.Message,), {
  'DESCRIPTOR' : _LEADERREGISTERREQUEST,
  '__module__' : 'leader_pb2'
  # @@protoc_insertion_point(class_scope:LeaderRegisterRequest)
  })
_sym_db.RegisterMessage(LeaderRegisterRequest)

LeaderRegisterResponse = _reflection.GeneratedProtocolMessageType('LeaderRegisterResponse', (_message.Message,), {
  'DESCRIPTOR' : _LEADERREGISTERRESPONSE,
  '__module__' : 'leader_pb2'
  # @@protoc_insertion_point(class_scope:LeaderRegisterResponse)
  })
_sym_db.RegisterMessage(LeaderRegisterResponse)

PollLeaderRequest = _reflection.GeneratedProtocolMessageType('PollLeaderRequest', (_message.Message,), {
  'DESCRIPTOR' : _POLLLEADERREQUEST,
  '__module__' : 'leader_pb2'
  # @@protoc_insertion_point(class_scope:PollLeaderRequest)
  })
_sym_db.RegisterMessage(PollLeaderRequest)

PollLeaderResponse = _reflection.GeneratedProtocolMessageType('PollLeaderResponse', (_message.Message,), {
  'DESCRIPTOR' : _POLLLEADERRESPONSE,
  '__module__' : 'leader_pb2'
  # @@protoc_insertion_point(class_scope:PollLeaderResponse)
  })
_sym_db.RegisterMessage(PollLeaderResponse)

PollLeaderJobRequest = _reflection.GeneratedProtocolMessageType('PollLeaderJobRequest', (_message.Message,), {
  'DESCRIPTOR' : _POLLLEADERJOBREQUEST,
  '__module__' : 'leader_pb2'
  # @@protoc_insertion_point(class_scope:PollLeaderJobRequest)
  })
_sym_db.RegisterMessage(PollLeaderJobRequest)

PollLeaderJobResponse = _reflection.GeneratedProtocolMessageType('PollLeaderJobResponse', (_message.Message,), {
  'DESCRIPTOR' : _POLLLEADERJOBRESPONSE,
  '__module__' : 'leader_pb2'
  # @@protoc_insertion_point(class_scope:PollLeaderJobResponse)
  })
_sym_db.RegisterMessage(PollLeaderJobResponse)

PipelineStageRequest = _reflection.GeneratedProtocolMessageType('PipelineStageRequest', (_message.Message,), {
  'DESCRIPTOR' : _PIPELINESTAGEREQUEST,
  '__module__' : 'leader_pb2'
  # @@protoc_insertion_point(class_scope:PipelineStageRequest)
  })
_sym_db.RegisterMessage(PipelineStageRequest)

PipelineStageResponse = _reflection.GeneratedProtocolMessageType('PipelineStageResponse', (_message.Message,), {
  'DESCRIPTOR' : _PIPELINESTAGERESPONSE,
  '__module__' : 'leader_pb2'
  # @@protoc_insertion_point(class_scope:PipelineStageResponse)
  })
_sym_db.RegisterMessage(PipelineStageResponse)

NewJobLeaderRequest = _reflection.GeneratedProtocolMessageType('NewJobLeaderRequest', (_message.Message,), {
  'DESCRIPTOR' : _NEWJOBLEADERREQUEST,
  '__module__' : 'leader_pb2'
  # @@protoc_insertion_point(class_scope:NewJobLeaderRequest)
  })
_sym_db.RegisterMessage(NewJobLeaderRequest)

NewJobLeaderResponse = _reflection.GeneratedProtocolMessageType('NewJobLeaderResponse', (_message.Message,), {
  'DESCRIPTOR' : _NEWJOBLEADERRESPONSE,
  '__module__' : 'leader_pb2'
  # @@protoc_insertion_point(class_scope:NewJobLeaderResponse)
  })
_sym_db.RegisterMessage(NewJobLeaderResponse)

LeaderHeartbeatRequest = _reflection.GeneratedProtocolMessageType('LeaderHeartbeatRequest', (_message.Message,), {
  'DESCRIPTOR' : _LEADERHEARTBEATREQUEST,
  '__module__' : 'leader_pb2'
  # @@protoc_insertion_point(class_scope:LeaderHeartbeatRequest)
  })
_sym_db.RegisterMessage(LeaderHeartbeatRequest)

LeaderHeartbeatResponse = _reflection.GeneratedProtocolMessageType('LeaderHeartbeatResponse', (_message.Message,), {
  'DESCRIPTOR' : _LEADERHEARTBEATRESPONSE,
  '__module__' : 'leader_pb2'
  # @@protoc_insertion_point(class_scope:LeaderHeartbeatResponse)
  })
_sym_db.RegisterMessage(LeaderHeartbeatResponse)

LeaderJobRequest = _reflection.GeneratedProtocolMessageType('LeaderJobRequest', (_message.Message,), {
  'DESCRIPTOR' : _LEADERJOBREQUEST,
  '__module__' : 'leader_pb2'
  # @@protoc_insertion_point(class_scope:LeaderJobRequest)
  })
_sym_db.RegisterMessage(LeaderJobRequest)

LeaderJobResponse = _reflection.GeneratedProtocolMessageType('LeaderJobResponse', (_message.Message,), {
  'DESCRIPTOR' : _LEADERJOBRESPONSE,
  '__module__' : 'leader_pb2'
  # @@protoc_insertion_point(class_scope:LeaderJobResponse)
  })
_sym_db.RegisterMessage(LeaderJobResponse)



_DISTRIBUTEDSERVER = _descriptor.ServiceDescriptor(
  name='DistributedServer',
  full_name='DistributedServer',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=1907,
  serialized_end=2584,
  methods=[
  _descriptor.MethodDescriptor(
    name='AddClient',
    full_name='DistributedServer.AddClient',
    index=0,
    containing_service=None,
    input_type=_WORKERREGISTERREQUEST,
    output_type=_WORKERREGISTERRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetUpdate',
    full_name='DistributedServer.GetUpdate',
    index=1,
    containing_service=None,
    input_type=_WORKERUPDATEREQUEST,
    output_type=_WORKERUPDATERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='RemoveClient',
    full_name='DistributedServer.RemoveClient',
    index=2,
    containing_service=None,
    input_type=_WORKERDEREGISTERREQUEST,
    output_type=_WORKERDEREGISTERRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='RegisterLeader',
    full_name='DistributedServer.RegisterLeader',
    index=3,
    containing_service=None,
    input_type=_LEADERREGISTERREQUEST,
    output_type=_LEADERREGISTERRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='PollLeader',
    full_name='DistributedServer.PollLeader',
    index=4,
    containing_service=None,
    input_type=_POLLLEADERREQUEST,
    output_type=_POLLLEADERRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='MarkStageCompleted',
    full_name='DistributedServer.MarkStageCompleted',
    index=5,
    containing_service=None,
    input_type=_PIPELINESTAGEREQUEST,
    output_type=_PIPELINESTAGERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='PollLeaderJob',
    full_name='DistributedServer.PollLeaderJob',
    index=6,
    containing_service=None,
    input_type=_POLLLEADERJOBREQUEST,
    output_type=_POLLLEADERJOBRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CreateNewJob',
    full_name='DistributedServer.CreateNewJob',
    index=7,
    containing_service=None,
    input_type=_NEWJOBLEADERREQUEST,
    output_type=_NEWJOBLEADERRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CheckLeaderHeartbeat',
    full_name='DistributedServer.CheckLeaderHeartbeat',
    index=8,
    containing_service=None,
    input_type=_LEADERHEARTBEATREQUEST,
    output_type=_LEADERHEARTBEATRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ShareCurrentJob',
    full_name='DistributedServer.ShareCurrentJob',
    index=9,
    containing_service=None,
    input_type=_LEADERJOBREQUEST,
    output_type=_LEADERJOBRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_DISTRIBUTEDSERVER)

DESCRIPTOR.services_by_name['DistributedServer'] = _DISTRIBUTEDSERVER

# @@protoc_insertion_point(module_scope)



import asyncio
import inspect
from typing import Dict, List, Union
from collections import defaultdict
from hedra.core.graphs.hooks.registry.registrar import registrar
from hedra.core.graphs.events import Event
from hedra.core.graphs.hooks.registry.registry_types import (
   ContextHook,
   EventHook,
   ValidateHook
)
from hedra.core.graphs.hooks.registry.registry_types.hook import Hook
from hedra.core.graphs.hooks.hook_types.hook_type import HookType
from hedra.core.graphs.hooks.hook_types.internal import Internal
from hedra.core.graphs.stages.types.stage_types import StageTypes
from hedra.core.graphs.stages.base.stage import Stage
from hedra.core.graphs.stages.validate.exceptions import ReservedMethodError

from .validators import Validator


class Validate(Stage):
    stage_type=StageTypes.VALIDATE
    
    def __init__(self) -> None:
        super().__init__()
        self.stages: Dict[StageTypes, Dict[str, Stage]] = {}
        self.validation_stages = {}
        self.hooks_by_name: Dict[str, Hook] = defaultdict(dict)

        base_stage_name = self.__class__.__name__
        self.logger.filesystem.sync['hedra.core'].info(f'{self.metadata_string} - Checking internal Hooks for stage - {base_stage_name}')

        for reserved_hook_name in self.internal_hooks:
            try:

                hook = registrar.reserved[base_stage_name].get(reserved_hook_name)

                assert hasattr(self, reserved_hook_name)
                assert isinstance(hook, Hook)
                assert hook.hook_type == HookType.INTERNAL

                internal_hook = getattr(self, hook.shortname)
                assert inspect.getsource(internal_hook) == inspect.getsource(hook.call)

            except AssertionError:
                raise ReservedMethodError(self, reserved_hook_name)

            self.logger.filesystem.sync['hedra.core'].info(f'{self.metadata_string} - Found internal Hook - {hook.name}:{hook.hook_id}- for stage - {base_stage_name}')
        
        for hook_type in HookType:
            self.hooks[hook_type] = []

        self.accepted_hook_types = [ HookType.VALIDATE, HookType.INTERNAL, HookType.EVENT, HookType.CONTEXT ]

    @Internal()
    async def run(self):
        events: List[Union[EventHook, Event]] = [event for event in self.hooks[HookType.EVENT] if hasattr(self, event.shortname)]
        pre_events: List[EventHook] = [
            event for event in events if isinstance(event, EventHook) and event.pre
        ]
        
        if len(pre_events) > 0:
            pre_event_names = ", ".join([
                event.shortname for event in pre_events
            ])

            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Executing PRE events - {pre_event_names}')
            await asyncio.wait([
                asyncio.create_task(event.call()) for event in pre_events
            ], timeout=self.stage_timeout)

        validator = Validator(self.stages, self.metadata_string)
        await validator.validate_stages()

        validator_hooks = await self.collect_validate_hooks()
        validator.add_hooks(HookType.VALIDATE, validator_hooks)

        await validator.validate_hooks()

        post_events: List[EventHook] = [
            event for event in events if isinstance(event, EventHook) and event.pre is False
        ]

        if len(post_events) > 0:
            post_event_names = ", ".join([
                event.shortname for event in post_events
            ])

            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Executing POST events - {post_event_names}')
            await asyncio.wait([
                asyncio.create_task(event.call()) for event in post_events
            ], timeout=self.stage_timeout)

        context_hooks: List[ContextHook] = self.hooks[HookType.CONTEXT]
        await asyncio.gather(*[
            asyncio.create_task(context_hook.call(self.context)) for context_hook in context_hooks
        ])

    async def collect_validate_hooks(self):       
        
        validate_hooks: List[ValidateHook] = []

        methods = inspect.getmembers(self, predicate=inspect.ismethod)
        for _, method in methods:
            method_name = method.__qualname__
            hook: Hook = registrar.all.get(method_name)
                        
            if hook and hook.hook_type is HookType.VALIDATE:
                validate_hooks.append(hook)
                await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Loaded Validation hook - {hook.name}:{hook.hook_id}:{hook.hook_id}')

        return validate_hooks


import functools
from .hook_type import HookType
from hedra.core.graphs.hooks.registry.registry_types.hook import Hook
from hedra.core.graphs.hooks.registry.registrar import registrar


@registrar(HookType.CONTEXT)
def context(key: str):
    
    def wrapper(func) -> Hook:

        @functools.wraps(func)
        def decorator(*args, **kwargs):
            return func(*args, **kwargs)

        return decorator

    return wrapper


from hedra.core.engines.types.websocket import WebsocketResult
from .http_event import HTTPEvent


class WebsocketEvent(HTTPEvent):

    def __init__(self, result: WebsocketResult) -> None:
        super(
            WebsocketEvent,
            self
        ).__init__(result)


import os
from hedra.core.graphs.stages import (
    Submit
)

from hedra.reporting.types import (
    CloudwatchConfig
)


class SubmitCloudwatchResultsStage(Submit):
    config=CloudwatchConfig(
        aws_access_key_id=os.getenv('AWS_ACCESS_KEY_ID', ''),
        aws_secret_access_key=os.getenv('AWS_SECRET_ACCESS_KEY', ''),
        region_name=os.getenv("AWS_REGION_NAME", ''),
        iam_role_arn=os.getenv('AWS_IAM_ROLE_ARN', ''),
        events_rule='events',
        metrics_rule='metrics',
        cloudwatch_source='<CLOUDWATCH_SOURCE>',
        cloudwatch_targets=[{
            'id': '<TARGET_ID>',
            'arn': '<TARGET_ARN>'
        }],
        aws_resource_arns=['<AWS_RESOURCE_ARN>']
    )


# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
from .leader_pb2 import (
        WorkerRegisterRequest,
        WorkerRegisterResponse,
        WorkerUpdateRequest,
        WorkerUpdateResponse,
        WorkerDeregisterRequest,
        WorkerDeregisterResponse,
        LeaderRegisterRequest,
        LeaderRegisterResponse,
        PollLeaderRequest,
        PollLeaderResponse,
        PipelineStageRequest,
        PipelineStageResponse,
        PollLeaderJobRequest,
        PollLeaderJobResponse,
        NewJobLeaderRequest,
        NewJobLeaderResponse,
        LeaderHeartbeatRequest,
        LeaderHeartbeatResponse,
        LeaderJobRequest,
        LeaderJobResponse
)



class DistributedServerStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.AddClient = channel.unary_unary(
                '/DistributedServer/AddClient',
                request_serializer=WorkerRegisterRequest.SerializeToString,
                response_deserializer=WorkerRegisterResponse.FromString,
                )
        self.GetUpdate = channel.unary_unary(
                '/DistributedServer/GetUpdate',
                request_serializer=WorkerUpdateRequest.SerializeToString,
                response_deserializer=WorkerUpdateResponse.FromString,
                )
        self.RemoveClient = channel.unary_unary(
                '/DistributedServer/RemoveClient',
                request_serializer=WorkerDeregisterRequest.SerializeToString,
                response_deserializer=WorkerDeregisterResponse.FromString,
                )
        self.RegisterLeader = channel.unary_unary(
                '/DistributedServer/RegisterLeader',
                request_serializer=LeaderRegisterRequest.SerializeToString,
                response_deserializer=LeaderRegisterResponse.FromString,
                )
        self.PollLeader = channel.unary_unary(
                '/DistributedServer/PollLeader',
                request_serializer=PollLeaderRequest.SerializeToString,
                response_deserializer=PollLeaderResponse.FromString,
                )
        self.MarkStageCompleted = channel.unary_unary(
                '/DistributedServer/MarkStageCompleted',
                request_serializer=PipelineStageRequest.SerializeToString,
                response_deserializer=PipelineStageResponse.FromString,
                )
        self.PollLeaderJob = channel.unary_unary(
                '/DistributedServer/PollLeaderJob',
                request_serializer=PollLeaderJobRequest.SerializeToString,
                response_deserializer=PollLeaderJobResponse.FromString,
                )
        self.CreateNewJob = channel.unary_unary(
                '/DistributedServer/CreateNewJob',
                request_serializer=NewJobLeaderRequest.SerializeToString,
                response_deserializer=NewJobLeaderResponse.FromString,
                )
        self.CheckLeaderHeartbeat = channel.unary_unary(
                '/DistributedServer/CheckLeaderHeartbeat',
                request_serializer=LeaderHeartbeatRequest.SerializeToString,
                response_deserializer=LeaderHeartbeatResponse.FromString,
                )
        self.ShareCurrentJob = channel.unary_unary(
                '/DistributedServer/ShareCurrentJob',
                request_serializer=LeaderJobRequest.SerializeToString,
                response_deserializer=LeaderJobResponse.FromString,
                )


class DistributedServerServicer(object):
    """Missing associated documentation comment in .proto file."""

    def AddClient(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetUpdate(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RemoveClient(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RegisterLeader(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def PollLeader(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def MarkStageCompleted(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def PollLeaderJob(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CreateNewJob(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CheckLeaderHeartbeat(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ShareCurrentJob(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_DistributedServerServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'AddClient': grpc.unary_unary_rpc_method_handler(
                    servicer.AddClient,
                    request_deserializer=WorkerRegisterRequest.FromString,
                    response_serializer=WorkerRegisterResponse.SerializeToString,
            ),
            'GetUpdate': grpc.unary_unary_rpc_method_handler(
                    servicer.GetUpdate,
                    request_deserializer=WorkerUpdateRequest.FromString,
                    response_serializer=WorkerUpdateResponse.SerializeToString,
            ),
            'RemoveClient': grpc.unary_unary_rpc_method_handler(
                    servicer.RemoveClient,
                    request_deserializer=WorkerDeregisterRequest.FromString,
                    response_serializer=WorkerDeregisterResponse.SerializeToString,
            ),
            'RegisterLeader': grpc.unary_unary_rpc_method_handler(
                    servicer.RegisterLeader,
                    request_deserializer=LeaderRegisterRequest.FromString,
                    response_serializer=LeaderRegisterResponse.SerializeToString,
            ),
            'PollLeader': grpc.unary_unary_rpc_method_handler(
                    servicer.PollLeader,
                    request_deserializer=PollLeaderRequest.FromString,
                    response_serializer=PollLeaderResponse.SerializeToString,
            ),
            'MarkStageCompleted': grpc.unary_unary_rpc_method_handler(
                    servicer.MarkStageCompleted,
                    request_deserializer=PipelineStageRequest.FromString,
                    response_serializer=PipelineStageResponse.SerializeToString,
            ),
            'PollLeaderJob': grpc.unary_unary_rpc_method_handler(
                    servicer.PollLeaderJob,
                    request_deserializer=PollLeaderJobRequest.FromString,
                    response_serializer=PollLeaderJobResponse.SerializeToString,
            ),
            'CreateNewJob': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateNewJob,
                    request_deserializer=NewJobLeaderRequest.FromString,
                    response_serializer=NewJobLeaderResponse.SerializeToString,
            ),
            'CheckLeaderHeartbeat': grpc.unary_unary_rpc_method_handler(
                    servicer.CheckLeaderHeartbeat,
                    request_deserializer=LeaderHeartbeatRequest.FromString,
                    response_serializer=LeaderHeartbeatResponse.SerializeToString,
            ),
            'ShareCurrentJob': grpc.unary_unary_rpc_method_handler(
                    servicer.ShareCurrentJob,
                    request_deserializer=LeaderJobRequest.FromString,
                    response_serializer=LeaderJobResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'DistributedServer', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class DistributedServer(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def AddClient(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/DistributedServer/AddClient',
            WorkerRegisterRequest.SerializeToString,
            WorkerRegisterResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetUpdate(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/DistributedServer/GetUpdate',
            WorkerUpdateRequest.SerializeToString,
            WorkerUpdateResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def RemoveClient(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/DistributedServer/RemoveClient',
            WorkerDeregisterRequest.SerializeToString,
            WorkerDeregisterResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def RegisterLeader(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/DistributedServer/RegisterLeader',
            LeaderRegisterRequest.SerializeToString,
            LeaderRegisterResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def PollLeader(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/DistributedServer/PollLeader',
            PollLeaderRequest.SerializeToString,
            PollLeaderResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def MarkStageCompleted(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/DistributedServer/MarkStageCompleted',
            PipelineStageRequest.SerializeToString,
            PipelineStageResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def PollLeaderJob(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/DistributedServer/PollLeaderJob',
            PollLeaderJobRequest.SerializeToString,
            PollLeaderJobResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CreateNewJob(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/DistributedServer/CreateNewJob',
            NewJobLeaderRequest.SerializeToString,
            NewJobLeaderResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CheckLeaderHeartbeat(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/DistributedServer/CheckLeaderHeartbeat',
            LeaderHeartbeatRequest.SerializeToString,
            LeaderHeartbeatResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ShareCurrentJob(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/DistributedServer/ShareCurrentJob',
            LeaderJobRequest.SerializeToString,
            LeaderJobResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)




import time
from numpy import float32, float64, int16, int32, int64
from typing import Any
from datetime import datetime


class AWSTimestreamErrorRecord:

    def __init__(self, record_name: str, record_stage: str, error_message: str, count: int, session_uuid: str) -> None:


        self.record_type = 'error_metric'
        self.record_name = record_name

        self.time = str(int(round(time.time() * 1000)))
        self.time_unit = 'MILLISECONDS'
        self.dimensions = [
            {
                "Name": "type",
                "Value": 'error_metric'
            },
            {
                "Name": "name",
                "Value": record_name
            },
            {
                "Name": "stage",
                "Value": record_stage
            },
            {
                "Name": "error_message",
                "Value": error_message
            },
            {
                "Name": "session_uuid",
                "Value": session_uuid
            }
        ]
        self.measure_name = f'{record_name}_errors_{session_uuid}'
        self.measure_value = str(count)
        self.measure_value_type = 'BIGINT'

    def to_dict(self):
        return  {
            "Time": self.time,
            "TimeUnit": self.time_unit,
            "Dimensions": self.dimensions,
            "MeasureName": self.measure_name,
            "MeasureValue": self.measure_value,
            "MeasureValueType": self.measure_value_type
        }



from weakref import ref
from asyncio import (
    Protocol,
    Transport
)
from asyncio.coroutines import iscoroutine
from hedra.core.engines.types.common.protocols.shared import FlowControlMixin
from hedra.core.engines.types.common.protocols.shared.reader import Reader
from hedra.core.engines.types.common.protocols.shared.writer import Writer


class UDPProtocol(FlowControlMixin, Protocol):

    _source_traceback = None

    def __init__(self, stream_reader, client_connected_cb=None, loop=None):
        super().__init__(loop=loop)

        if stream_reader is not None:
            self._stream_reader_wr: Reader = ref(stream_reader)
            self._source_traceback = stream_reader._source_traceback
        else:
            self._stream_reader_wr = None
        if client_connected_cb is not None:
            # This is a stream created by the `create_server()` function.
            # Keep a strong reference to the reader until a connection
            # is established.
            self._strong_reader = stream_reader
        self._reject_connection = False
        self._stream_writer: Writer = None
        self._transport: Transport = None
        self._client_connected_cb = client_connected_cb
        
        self._closed = self._loop.create_future()

    @property
    def _stream_reader(self) -> Reader:
        if self._stream_reader_wr is None:
            return None
        return self._stream_reader_wr()

    def _replace_writer(self, writer: Writer):
        transport = writer.transport
        self._stream_writer = writer
        self._transport = transport

    def connection_made(self, transport: Transport):
        if self._reject_connection:
            context = {
                'message': ('An open stream was garbage collected prior to '
                            'establishing network connection; '
                            'call "stream.close()" explicitly.')
            }
            if self._source_traceback:
                context['source_traceback'] = self._source_traceback
            self._loop.call_exception_handler(context)
            transport.abort()
            return
        self._transport = transport
        reader: Reader = self._stream_reader
        if reader is not None:
            reader.set_transport(transport)

        if self._client_connected_cb is not None:
            self._stream_writer = Reader(transport, self,
                                               reader,
                                               self._loop)
            res = self._client_connected_cb(reader,
                                            self._stream_writer)
            if iscoroutine(res):
                self._loop.create_task(res)
            self._strong_reader = None

    def connection_lost(self, exc):
        reader: Reader = self._stream_reader
        if reader is not None:
            if exc is None:
                reader.feed_eof()
            else:
                reader.set_exception(exc)
        if not self._closed.done():
            if exc is None:
                self._closed.set_result(None)
            else:
                self._closed.set_exception(exc)
        super().connection_lost(exc)
        self._stream_reader_wr = None
        self._stream_writer = None
        self._transport = None

    def error_received(self, exc):
        raise exc

    def datagram_received(self, data):
        reader = self._stream_reader
        if reader is not None:
            reader.feed_data(data)

    def eof_received(self):
        reader: Reader = self._stream_reader
        if reader is not None:
            reader.feed_eof()

        return True

    def _get_close_waiter(self, stream):
        return self._closed

    def __del__(self):
        # Prevent reports about unhandled exceptions.
        # Better than self._closed._log_traceback = False hack
        try:
            closed = self._closed
        except AttributeError:
            pass  # failed constructor
        else:
            if closed.done() and not closed.cancelled():
                closed.exception()


import time
import asyncio
from typing import AsyncIterable, Dict, List, Union
from hedra.core.graphs.hooks.hook_types.hook_type import HookType
from hedra.core.graphs.hooks.registry.registry_types import ActionHook, TaskHook
from hedra.plugins.types.persona import (
    PersonaPlugin,
    setup,
    generate,
    shutdown
)


class CustomPersona(PersonaPlugin):

    @setup()
    async def setup(self, hooks: Dict[HookType, List[Union[ActionHook, TaskHook]]]):
        return super().setup(hooks)

    @generate()
    async def generate_next(self) -> AsyncIterable[int]:
        total_time = self.total_time
        elapsed = 0
        idx = 0

        start = time.time()

        while elapsed < total_time:
            yield idx%self.actions_count
            await asyncio.sleep(0)
            
            idx += 1
            

        elapsed = time.time() - start

    @shutdown()
    async def shutdown(self):
        pass


from ctypes import Union
from typing import List
import uuid
from .leader_service import LeaderService

class LeaderRegistry:


    def __init__(self) -> None:
        self.registered = []
        self.elected_leader = None
        self.expected_leaders = []
        self.expected_leaders_count = 0
        self.registered_count = 1
        self._registered = set()
        self.leader_id = None

    async def setup(self, services: List[LeaderService]):
        self.registered = [service for service in services if service.host == True]
        self.leader_id = self.registered[0].service_id

        for leader in self.registered:
            self._registered.add(leader.service_address)

        self.expected_leaders = [service for service in services if service.host == False]
        self.expected_leaders_count = len(self.expected_leaders) + self.registered_count

    async def use_dynamic_registration(self):
        return self.expected_leaders_count > 0 and len(self.expected_leaders) == 0

    async def set_elected_leader(self, elected_leader):
        self.elected_leader = elected_leader

    async def remove_service(self, removed_address):
        self.expected_leaders = [service for service in self.expected_leaders if service.service_address != removed_address]
        if self.elected_leader.service_address == removed_address:
            self.elected_leader = None

    async def registration_complete(self):
        return self.registered_count >= self.expected_leaders_count

    async def services(self):
        for service in self.expected_leaders:
            yield service

    async def register(self, leader_registration_request):
        leader_id = uuid.UUID(leader_registration_request.host_id, version=4)        
        self.registered += [ LeaderService(leader_registration_request.host_address, leader_id) ]
        
        if leader_registration_request.host_address not in self._registered:
            self._registered.add(leader_registration_request.host_address)
            self.registered_count += 1

    async def register_by_response(self, leader_registration_response):
        leader_id = uuid.UUID(leader_registration_response.host_id, version=4)
        self.registered += [ LeaderService(leader_registration_response.host_address, leader_id) ]

        if leader_registration_response.host_address not in self._registered:
            self._registered.add(leader_registration_response.host_address)
            self.registered_count += 1


    async def find_min_id(self):        
        return min(self.registered, key=lambda registered_leader: registered_leader.service_id.int)

    async def share_job(self, job):
        pipeline_updates = []

        for leader in self.registered:
            if leader.host is False:
                updated_pipeline_status = await leader.share_current_job_config(job)

                pipeline_updates += [updated_pipeline_status]

        return pipeline_updates

    async def registered_count_equals_expected_count(self):
        return self.registered_count >= self.expected_leaders_count


from .reporter_plugin import ReporterPlugin
from .reporter_config import ReporterConfig
from .reporter_metrics import Metrics
from .hooks.types import (
    process_events,
    process_shared,
    process_metrics,
    process_custom,
    process_errors,
    reporter_connect,
    reporter_close
)


from .analyze import Analyze
from .checkpoint import Checkpoint
from .execute import Execute
from .optimize import Optimize
from .setup import Setup
from .validate import Validate
from .teardown import Teardown
from .complete import Complete
from .submit import Submit
from .error import Error
from .idle import Idle
from .wait import Wait


from typing import Dict, Any, Awaitable, Union
from hedra.core.engines.client.config import Config
from hedra.core.engines.types.common.types import RequestTypes
from hedra.core.graphs.hooks.registry.registry_types import ActionHook, TaskHook
from hedra.core.personas.types.default_persona import DefaultPersona
from .assemble_graphql_action import assemble_graphql_action
from .assemble_graphql_http2_action import assemble_graphql_http2_action
from .assemble_grpc_action import assemble_grpc_action
from .assemble_http_action import assemble_http_action
from .assemble_http2_action import assemble_http2_action
from .assemble_playwright_action import assemble_playwright_command
from .assemble_udp_action import assemble_udp_action
from .assemble_plugin_action import assemble_plugin_action
from .assemble_task import assemble_task
from .assemble_websocket_action import assemble_websocket_action


class ActionAssembler:

    def __init__(
        self,
        hook: Union[ActionHook, TaskHook],
        hook_action: Dict[str, Any],
        persona: DefaultPersona,
        config: Config,
        metadata_string: str
    ) -> None:
        self.hook = hook
        self.hook_action = hook_action
        self.persona = persona
        self.config = config

        self._assembler_types = {
            RequestTypes.GRAPHQL: assemble_graphql_action,
            RequestTypes.GRAPHQL_HTTP2: assemble_graphql_http2_action,
            RequestTypes.GRPC: assemble_grpc_action,
            RequestTypes.HTTP: assemble_http_action,
            RequestTypes.HTTP2: assemble_http2_action,
            RequestTypes.PLAYWRIGHT: assemble_playwright_command,
            RequestTypes.TASK: assemble_task,
            RequestTypes.UDP: assemble_udp_action,
            RequestTypes.WEBSOCKET: assemble_websocket_action
        }

        self.metadata_string = metadata_string

    def assemble(self, action_type: RequestTypes) -> Awaitable[Union[ActionHook, TaskHook]]:
        return self._assembler_types.get(
            action_type,
            assemble_plugin_action
        )(
            self.hook,
            self.hook_action,
            self.persona,
            self.config,
            self.metadata_string
        )


import datetime
from typing import List
from .action import RepoAction
from .config import RepoConfig



class CreateGitignore(RepoAction):

    def __init__(self, config: RepoConfig, graph_files: List[str]) -> None:
        super(
            CreateGitignore,
            self
        ).__init__(config, graph_files)

    def execute(self):

        self.logger.hedra.sync.debug(f'CreateGitignore: {self.action_id} - creating .gitignore at path - {self.config.path}')
        self._setup()
        self._update_ignore(force_create=True)

        current_time = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%SZ')
        gitignore_mmessage = f"Hedra graph update: {self.config.path} - {current_time}" 


        self.logger.hedra.sync.debug(f'CreateGitignore: {self.action_id} - making local commit to add new .gitignore at path - {self.config.path}')
        self.repo.index.commit(gitignore_mmessage)


import json
import dill
import threading
import os
import time
from typing import Dict, Any, List, Union
from hedra.core.engines.client.config import Config
from hedra.core.graphs.hooks.registry.registry_types import ActionHook, TaskHook
from hedra.core.graphs.hooks.hook_types.hook_type import HookType
from hedra.core.engines.types.playwright import MercuryPlaywrightClient, ContextConfig
from hedra.core.engines.types.registry import RequestTypes
from hedra.core.engines.types.registry import registered_engines
from hedra.core.personas.persona_registry import registered_personas
from hedra.plugins.types.plugin_types import PluginType
from hedra.plugins.types.engine.engine_plugin import EnginePlugin
from hedra.plugins.types.persona.persona_plugin import PersonaPlugin
from hedra.core.graphs.stages.base.import_tools import import_stages, import_plugins
from hedra.core.personas import get_persona
from hedra.logging import (
    HedraLogger,
    LoggerTypes,
    logging_manager
)

from hedra.core.graphs.stages.base.parallel.partition_method import PartitionMethod
from .action_assembly import ActionAssembler

async def start_execution(parallel_config: Dict[str, Any]):

    hedra_config_filepath = os.path.join(
        os.getcwd(),
        '.hedra.json'
    )

    hedra_config = {}
    if os.path.exists(hedra_config_filepath):
        with open(hedra_config_filepath, 'r') as hedra_config_file:
            hedra_config = json.load(hedra_config_file)

    logging_config = hedra_config.get('logging', {})
    logfiles_directory = logging_config.get(
        'logfiles_directory',
        os.getcwd()
    )

    log_level = logging_config.get('log_level', 'info')

    logging_manager.disable(
        LoggerTypes.DISTRIBUTED,
        LoggerTypes.DISTRIBUTED_FILESYSTEM
    )

    logging_manager.update_log_level(log_level)
    logging_manager.logfiles_directory = logfiles_directory


    logger = HedraLogger()
    logger.initialize()
    await logger.filesystem.aio.create_logfile('hedra.core.log')
    logger.filesystem.create_filelogger('hedra.core.log')

    start = time.monotonic()

    graph_name = parallel_config.get('graph_name')
    graph_path: str= parallel_config.get('graph_path')
    graph_id = parallel_config.get('graph_id')
    source_stage_name = parallel_config.get('source_stage_name')
    source_stage_plugins = parallel_config.get('source_stage_plugins')
    source_stage_id = parallel_config.get('source_stage_id')
    thread_id = threading.current_thread().ident
    process_id = os.getpid()

    metadata_string = f'Graph - {graph_name}:{graph_id} - thread:{thread_id} - process:{process_id} - Stage: {source_stage_name}:{source_stage_id} - '

    execution_hooks = parallel_config.get('hooks')
    partition_method = parallel_config.get('partition_method')
    persona_config: Config = parallel_config.get('config')
    workers = parallel_config.get('workers')
    worker_id = parallel_config.get('worker_id')

    execution_hooks_count = len(execution_hooks)

    if partition_method == PartitionMethod.BATCHES and persona_config.optimized is False:
        if workers == worker_id:
            persona_config.batch_size = int(persona_config.batch_size/workers) + (persona_config.batch_size%workers)
        
        else:
            persona_config.batch_size = int(persona_config.batch_size/workers)

    plugins_by_type = import_plugins(graph_path)

    stage_persona_plugins: List[str] = source_stage_plugins[PluginType.PERSONA]
    persona_plugins: Dict[str, PersonaPlugin] = plugins_by_type[PluginType.PERSONA]

    stage_engine_plugins: List[str] = source_stage_plugins[PluginType.ENGINE]
    engine_plugins: Dict[str, EnginePlugin] = plugins_by_type[PluginType.ENGINE]
    
    for plugin_name in stage_persona_plugins:
        plugin = persona_plugins.get(plugin_name)
        plugin.name = plugin_name
        registered_personas[plugin_name] = lambda config: plugin(config)
    
    for plugin_name in stage_engine_plugins:
        plugin = engine_plugins.get(plugin_name)
        plugin.name = plugin_name
        registered_engines[plugin_name] = lambda config: plugin(config)

    persona = get_persona(persona_config)
    persona.workers = workers


    await logger.filesystem.aio['hedra.core'].info(
        f'{metadata_string} - Executing {execution_hooks_count} actions with a batch size of {persona_config.batch_size} for {persona_config.total_time} seconds using Persona - {persona.type.capitalize()}'
    )

    hooks: Dict[HookType, List[Union[ActionHook, TaskHook]]] = {
        HookType.ACTION: [],
        HookType.TASK: [],
    }
    
    actions = {}
    for hook_action in execution_hooks:
        hook_type = hook_action.get('hook_type', HookType.ACTION)

        if hook_type == HookType.ACTION:

            hook = ActionHook(
                hook_action.get('hook_name'),
                hook_action.get('hook_shortname'),
                None
            )

        else:
            
            hook = TaskHook(
                hook_action.get('hook_name'),
                hook_action.get('hook_shortname'),
                None
            )

        hook.stage = hook_action.get('stage')

        action_type: str = hook_action.get('type')
        action_assembler = ActionAssembler(
            hook,
            hook_action,
            persona,
            persona_config,
            metadata_string
        )

        assembled_hook = await action_assembler.assemble(action_type)
        actions[assembled_hook.name] = assembled_hook.action
        
        await logger.filesystem.aio['hedra.core'].info(
            f'{metadata_string} - Assembled hook - {hook.name}:{hook.hook_id} - using {action_type.capitalize()} Engine'
        )

        hooks[hook_type].append(assembled_hook)

    actions_and_tasks = [
        *hooks.get(HookType.ACTION, []),
        *hooks.get(HookType.TASK, [])
    ]

    for hook in actions_and_tasks:
        if hook.action.hooks.notify:
            for idx, listener_name in enumerate(hook.action.hooks.listeners):
                hook.action.hooks.listeners[idx] = actions.get(listener_name)

    persona.setup(hooks, metadata_string)

    if action_type == RequestTypes.PLAYWRIGHT and isinstance(hook.session, MercuryPlaywrightClient):

            await logger.filesystem.aio['hedra.core'].info(f'{metadata_string} - Setting up Playwright Session')

            await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - Playwright Session - {hook.session.session_id} - Browser Type: {persona_config.browser_type}')
            await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - Playwright Session - {hook.session.session_id} - Device Type: {persona_config.device_type}')
            await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - Playwright Session - {hook.session.session_id} - Locale: {persona_config.locale}')
            await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - Playwright Session - {hook.session.session_id} - geolocation: {persona_config.geolocation}')
            await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - Playwright Session - {hook.session.session_id} - Permissions: {persona_config.permissions}')
            await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - Playwright Session - {hook.session.session_id} - Color Scheme: {persona_config.color_scheme}')

            await hook.session.setup(ContextConfig(
                browser_type=persona_config.browser_type,
                device_type=persona_config.device_type,
                locale=persona_config.locale,
                geolocation=persona_config.geolocation,
                permissions=persona_config.permissions,
                color_scheme=persona_config.color_scheme,
                options=persona_config.playwright_options
            ))


    await logger.filesystem.aio['hedra.core'].info(f'{metadata_string} - Starting execution')

    results = await persona.execute()

    elapsed = time.monotonic() - start

    await logger.filesystem.aio['hedra.core'].info(f'{metadata_string} - Execution complete - Time (including addtional setup) took: {round(elapsed, 2)} seconds')

    return {
        'results': results,
        'total_results': len(results),
        'total_elapsed': persona.total_elapsed
    }


def execute_actions(parallel_config: str):

    try:
        import asyncio
        import uvloop
        uvloop.install()
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        
        parallel_config: Dict[str, Any] = dill.loads(parallel_config)

        return loop.run_until_complete(
            start_execution(parallel_config)
        )

    except Exception as e:
        raise e


from pydantic import BaseModel
from hedra.reporting.types.common.types import ReporterTypes


class AWSLambdaConfig(BaseModel):
    aws_access_key_id: str
    aws_secret_access_key: str
    region_name: str
    events_lambda: str
    metrics_lambda: str
    reporter_type: ReporterTypes=ReporterTypes.AWSLambda


# -*- coding: utf-8 -*-
"""
hpack/hpack
~~~~~~~~~~~

Implements the HPACK header compression algorithm as detailed by the IETF.
"""
import logging

from hpack.table import HeaderTable, table_entry_size
from hpack.exceptions import (
    HPACKDecodingError, OversizedHeaderListError, InvalidTableSizeError
)
from hpack.huffman import HuffmanEncoder
from hpack.huffman_constants import (
    REQUEST_CODES, REQUEST_CODES_LENGTH
)
from hpack.huffman_table import decode_huffman
from hpack.struct import HeaderTuple, NeverIndexedHeaderTuple

log = logging.getLogger(__name__)

INDEX_NONE = b'\x00'
INDEX_NEVER = b'\x10'
INDEX_INCREMENTAL = b'\x40'

# Precompute 2^i for 1-8 for use in prefix calcs.
# Zero index is not used but there to save a subtraction
# as prefix numbers are not zero indexed.
_PREFIX_BIT_MAX_NUMBERS = [(2 ** i) - 1 for i in range(9)]

basestring = (str, bytes)


# We default the maximum header list we're willing to accept to 64kB. That's a
# lot of headers, but if applications want to raise it they can do.
DEFAULT_MAX_HEADER_LIST_SIZE = 2 ** 16


def _unicode_if_needed(header, raw):
    """
    Provides a header as a unicode string if raw is False, otherwise returns
    it as a bytestring.
    """
    name = bytes(header[0])
    value = bytes(header[1])
    if not raw:
        name = name.decode('utf-8')
        value = value.decode('utf-8')
    return header.__class__(name, value)


def encode_integer(integer, prefix_bits):
    """
    This encodes an integer according to the wacky integer encoding rules
    defined in the HPACK spec.
    """
    log.debug("Encoding %d with %d bits", integer, prefix_bits)

    if integer < 0:
        raise ValueError(
            "Can only encode positive integers, got %s" % integer
        )

    if prefix_bits < 1 or prefix_bits > 8:
        raise ValueError(
            "Prefix bits must be between 1 and 8, got %s" % prefix_bits
        )

    max_number = _PREFIX_BIT_MAX_NUMBERS[prefix_bits]

    if integer < max_number:
        return bytearray([integer])  # Seriously?
    else:
        elements = [max_number]
        integer -= max_number

        while integer >= 128:
            elements.append((integer & 127) + 128)
            integer >>= 7

        elements.append(integer)

        return bytearray(elements)


def decode_integer(data, prefix_bits):
    """
    This decodes an integer according to the wacky integer encoding rules
    defined in the HPACK spec. Returns a tuple of the decoded integer and the
    number of bytes that were consumed from ``data`` in order to get that
    integer.
    """
    if prefix_bits < 1 or prefix_bits > 8:
        raise ValueError(
            "Prefix bits must be between 1 and 8, got %s" % prefix_bits
        )

    max_number = _PREFIX_BIT_MAX_NUMBERS[prefix_bits]
    index = 1
    shift = 0
    mask = (0xFF >> (8 - prefix_bits))

    try:
        number = data[0] & mask
        if number == max_number:
            while True:
                next_byte = data[index]
                index += 1

                if next_byte >= 128:
                    number += (next_byte - 128) << shift
                else:
                    number += next_byte << shift
                    break
                shift += 7

    except IndexError:
        raise HPACKDecodingError(
            "Unable to decode HPACK integer representation from %r" % data
        )

    log.debug("Decoded %d, consumed %d bytes", number, index)

    return number, index



def _to_bytes(string):
    """
    Convert string to bytes.
    """
    if not isinstance(string, basestring):  # pragma: no cover
        string = str(string)

    return string if isinstance(string, bytes) else string.encode('utf-8')



class Decoder:
    """
    An HPACK decoder object.

    .. versionchanged:: 2.3.0
       Added ``max_header_list_size`` argument.

    :param max_header_list_size: The maximum decompressed size we will allow
        for any single header block. This is a protection against DoS attacks
        that attempt to force the application to expand a relatively small
        amount of data into a really large header list, allowing enormous
        amounts of memory to be allocated.

        If this amount of data is exceeded, a `OversizedHeaderListError
        <hpack.OversizedHeaderListError>` exception will be raised. At this
        point the connection should be shut down, as the HPACK state will no
        longer be usable.

        Defaults to 64kB.
    :type max_header_list_size: ``int``
    """

    __slots__ = (
        'header_table', 
        'max_header_list_size',
        'max_allowed_table_size'
    )

    def __init__(self, max_header_list_size=DEFAULT_MAX_HEADER_LIST_SIZE):
        self.header_table = HeaderTable()

        #: The maximum decompressed size we will allow for any single header
        #: block. This is a protection against DoS attacks that attempt to
        #: force the application to expand a relatively small amount of data
        #: into a really large header list, allowing enormous amounts of memory
        #: to be allocated.
        #:
        #: If this amount of data is exceeded, a `OversizedHeaderListError
        #: <hpack.OversizedHeaderListError>` exception will be raised. At this
        #: point the connection should be shut down, as the HPACK state will no
        #: longer be usable.
        #:
        #: Defaults to 64kB.
        #:
        #: .. versionadded:: 2.3.0
        self.max_header_list_size = max_header_list_size

        #: Maximum allowed header table size.
        #:
        #: A HTTP/2 implementation should set this to the most recent value of
        #: SETTINGS_HEADER_TABLE_SIZE that it sent *and has received an ACK
        #: for*. Once this setting is set, the actual header table size will be
        #: checked at the end of each decoding run and whenever it is changed,
        #: to confirm that it fits in this size.
        self.max_allowed_table_size = self.header_table.maxsize

    @property
    def header_table_size(self):
        """
        Controls the size of the HPACK header table.
        """
        return self.header_table.maxsize

    @header_table_size.setter
    def header_table_size(self, value):
        self.header_table.maxsize = value

    def decode(self, data, raw=False):
        """
        Takes an HPACK-encoded header block and decodes it into a header set.

        :param data: A bytestring representing a complete HPACK-encoded header
                     block.
        :param raw: (optional) Whether to return the headers as tuples of raw
                    byte strings or to decode them as UTF-8 before returning
                    them. The default value is False, which returns tuples of
                    Unicode strings
        :returns: A list of two-tuples of ``(name, value)`` representing the
                  HPACK-encoded headers, in the order they were decoded.
        :raises HPACKDecodingError: If an error is encountered while decoding
                                    the header block.
        """
        log.debug("Decoding %s", data)

        data_mem = bytearray(data)
        headers = []
        data_len = len(data)
        inflated_size = 0
        current_index = 0

        while current_index < data_len:
            # Work out what kind of header we're decoding.
            # If the high bit is 1, it's an indexed field.
            current = data[current_index]
            indexed = True if current & 0x80 else False

            # Otherwise, if the second-highest bit is 1 it's a field that does
            # alter the header table.
            literal_index = True if current & 0x40 else False

            # Otherwise, if the third-highest bit is 1 it's an encoding context
            # update.
            encoding_update = True if current & 0x20 else False

            if indexed:
                header, consumed = self._decode_indexed(
                    data_mem[current_index:]
                )
            elif literal_index:
                # It's a literal header that does affect the header table.
                header, consumed = self._decode_literal_index(
                    data_mem[current_index:]
                )
            elif encoding_update:
                # It's an update to the encoding context. These are forbidden
                # in a header block after any actual header.
                if headers:
                    raise HPACKDecodingError(
                        "Table size update not at the start of the block"
                    )
                consumed = self._update_encoding_context(
                    data_mem[current_index:]
                )
                header = None
            else:
                # It's a literal header that does not affect the header table.
                header, consumed = self._decode_literal_no_index(
                    data_mem[current_index:]
                )

            if header:
                headers.append(header)
                inflated_size += table_entry_size(*header)

                if inflated_size > self.max_header_list_size:
                    raise OversizedHeaderListError(
                        "A header list larger than %d has been received" %
                        self.max_header_list_size
                    )

            current_index += consumed

        # Confirm that the table size is lower than the maximum. We do this
        # here to ensure that we catch when the max has been *shrunk* and the
        # remote peer hasn't actually done that.
        self._assert_valid_table_size()

        try:
            return [_unicode_if_needed(h, raw) for h in headers]
        except UnicodeDecodeError:
            raise HPACKDecodingError("Unable to decode headers as UTF-8.")

    def _assert_valid_table_size(self):
        """
        Check that the table size set by the encoder is lower than the maximum
        we expect to have.
        """
        if self.header_table_size > self.max_allowed_table_size:
            raise InvalidTableSizeError(
                "Encoder did not shrink table size to within the max"
            )

    def _update_encoding_context(self, data):
        """
        Handles a byte that updates the encoding context.
        """
        # We've been asked to resize the header table.
        new_size, consumed = decode_integer(data, 5)
        if new_size > self.max_allowed_table_size:
            raise InvalidTableSizeError(
                "Encoder exceeded max allowable table size"
            )
        self.header_table_size = new_size
        return consumed

    def _decode_indexed(self, data):
        """
        Decodes a header represented using the indexed representation.
        """
        index, consumed = decode_integer(data, 7)
        header = HeaderTuple(*self.header_table.get_by_index(index))
        log.debug("Decoded %s, consumed %d", header, consumed)
        return header, consumed

    def _decode_literal_no_index(self, data):
        return self._decode_literal(data, False)

    def _decode_literal_index(self, data):
        return self._decode_literal(data, True)

    def _decode_literal(self, data, should_index):
        """
        Decodes a header represented with a literal.
        """
        total_consumed = 0

        # When should_index is true, if the low six bits of the first byte are
        # nonzero, the header name is indexed.
        # When should_index is false, if the low four bits of the first byte
        # are nonzero the header name is indexed.
        if should_index:
            indexed_name = data[0] & 0x3F
            name_len = 6
            not_indexable = False
        else:
            high_byte = data[0]
            indexed_name = high_byte & 0x0F
            name_len = 4
            not_indexable = high_byte & 0x10

        if indexed_name:
            # Indexed header name.
            index, consumed = decode_integer(data, name_len)
            name = self.header_table.get_by_index(index)[0]

            total_consumed = consumed
            length = 0
        else:
            # Literal header name. The first byte was consumed, so we need to
            # move forward.
            data = data[1:]

            length, consumed = decode_integer(data, 7)
            name = data[consumed:consumed + length]
            if len(name) != length:
                raise HPACKDecodingError("Truncated header block")

            if data[0] & 0x80:
                name = decode_huffman(name)
            total_consumed = consumed + length + 1  # Since we moved forward 1.

        data = data[consumed + length:]

        # The header value is definitely length-based.
        length, consumed = decode_integer(data, 7)
        value = data[consumed:consumed + length]
        if len(value) != length:
            raise HPACKDecodingError("Truncated header block")

        if data[0] & 0x80:
            value = decode_huffman(value)

        # Updated the total consumed length.
        total_consumed += length + consumed

        # If we have been told never to index the header field, encode that in
        # the tuple we use.
        if not_indexable:
            header = NeverIndexedHeaderTuple(name, value)
        else:
            header = HeaderTuple(name, value)

        # If we've been asked to index this, add it to the header table.
        if should_index:
            self.header_table.add(name, value)

        log.debug(
            "Decoded %s, total consumed %d bytes, indexed %s",
            header,
            total_consumed,
            should_index
        )

        return header, total_consumed



from enum import Enum


class StageStates(Enum):
    INITIALIZED='INITIALIZED'
    SETTING_UP='SETTING_UP'
    SETUP='SETUP'
    VALIDATING='VALIDATING'
    VALIDATED='VALIDATED'
    OPTIMIZING='OPTIMIZING'
    OPTIMIZED='OPTIMIZED'
    EXECUTING='EXECUTING'
    EXECUTED='EXECUTED'
    ANALYZING='ANALYZING'
    ANALYZED='ANALYZED'
    CHECKPOINTING='CHECKPOINTING'
    CHECKPOINTED='CHECKPOINTED'
    SUBMITTING='SUBMITTING'
    SUBMITTED='SUBMITTED'
    TEARDOWN_INITIALIZED='TEARDOWN_INITIALIZED'
    TEARDOWN_COMPLETE='TEARDOWN_COMPLETE'
    COMPLETE='COMPLETE'
    ERRORED='ERRORED'


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: bootstrap.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='bootstrap.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x0f\x62ootstrap.proto\"{\n\x16ServiceRegisterRequest\x12\x17\n\x0fservice_address\x18\x01 \x01(\t\x12\x12\n\nservice_id\x18\x02 \x01(\t\x12\x1e\n\x16service_server_address\x18\x03 \x01(\t\x12\x14\n\x0cservice_name\x18\x04 \x01(\t\"e\n ServiceRegistrationStatusRequest\x12\x17\n\x0fservice_address\x18\x01 \x01(\t\x12\x12\n\nservice_id\x18\x02 \x01(\t\x12\x14\n\x0cservice_name\x18\x03 \x01(\t\"\xe8\x01\n!ServiceRegistrationStatusResponse\x12!\n\x19\x62ootstrap_service_address\x18\x01 \x01(\t\x12\x1c\n\x14\x62ootstrap_service_id\x18\x02 \x01(\t\x12\x1c\n\x14registered_addresses\x18\x03 \x03(\t\x12\x16\n\x0eregistered_ids\x18\x04 \x03(\t\x12#\n\x1bregistered_server_addresses\x18\x05 \x03(\t\x12\x18\n\x10registered_count\x18\x06 \x01(\x05\x12\r\n\x05ready\x18\x07 \x01(\x08\x32\xcc\x01\n\x0f\x42ootstrapServer\x12P\n\x0fRegisterService\x12\x17.ServiceRegisterRequest\x1a\".ServiceRegistrationStatusResponse\"\x00\x12g\n\x1cGetServiceRegistrationStatus\x12!.ServiceRegistrationStatusRequest\x1a\".ServiceRegistrationStatusResponse\"\x00\x62\x06proto3'
)




_SERVICEREGISTERREQUEST = _descriptor.Descriptor(
  name='ServiceRegisterRequest',
  full_name='ServiceRegisterRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='service_address', full_name='ServiceRegisterRequest.service_address', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='service_id', full_name='ServiceRegisterRequest.service_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='service_server_address', full_name='ServiceRegisterRequest.service_server_address', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='service_name', full_name='ServiceRegisterRequest.service_name', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=19,
  serialized_end=142,
)


_SERVICEREGISTRATIONSTATUSREQUEST = _descriptor.Descriptor(
  name='ServiceRegistrationStatusRequest',
  full_name='ServiceRegistrationStatusRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='service_address', full_name='ServiceRegistrationStatusRequest.service_address', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='service_id', full_name='ServiceRegistrationStatusRequest.service_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='service_name', full_name='ServiceRegistrationStatusRequest.service_name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=144,
  serialized_end=245,
)


_SERVICEREGISTRATIONSTATUSRESPONSE = _descriptor.Descriptor(
  name='ServiceRegistrationStatusResponse',
  full_name='ServiceRegistrationStatusResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='bootstrap_service_address', full_name='ServiceRegistrationStatusResponse.bootstrap_service_address', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='bootstrap_service_id', full_name='ServiceRegistrationStatusResponse.bootstrap_service_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='registered_addresses', full_name='ServiceRegistrationStatusResponse.registered_addresses', index=2,
      number=3, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='registered_ids', full_name='ServiceRegistrationStatusResponse.registered_ids', index=3,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='registered_server_addresses', full_name='ServiceRegistrationStatusResponse.registered_server_addresses', index=4,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='registered_count', full_name='ServiceRegistrationStatusResponse.registered_count', index=5,
      number=6, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ready', full_name='ServiceRegistrationStatusResponse.ready', index=6,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=248,
  serialized_end=480,
)

DESCRIPTOR.message_types_by_name['ServiceRegisterRequest'] = _SERVICEREGISTERREQUEST
DESCRIPTOR.message_types_by_name['ServiceRegistrationStatusRequest'] = _SERVICEREGISTRATIONSTATUSREQUEST
DESCRIPTOR.message_types_by_name['ServiceRegistrationStatusResponse'] = _SERVICEREGISTRATIONSTATUSRESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ServiceRegisterRequest = _reflection.GeneratedProtocolMessageType('ServiceRegisterRequest', (_message.Message,), {
  'DESCRIPTOR' : _SERVICEREGISTERREQUEST,
  '__module__' : 'bootstrap_pb2'
  # @@protoc_insertion_point(class_scope:ServiceRegisterRequest)
  })
_sym_db.RegisterMessage(ServiceRegisterRequest)

ServiceRegistrationStatusRequest = _reflection.GeneratedProtocolMessageType('ServiceRegistrationStatusRequest', (_message.Message,), {
  'DESCRIPTOR' : _SERVICEREGISTRATIONSTATUSREQUEST,
  '__module__' : 'bootstrap_pb2'
  # @@protoc_insertion_point(class_scope:ServiceRegistrationStatusRequest)
  })
_sym_db.RegisterMessage(ServiceRegistrationStatusRequest)

ServiceRegistrationStatusResponse = _reflection.GeneratedProtocolMessageType('ServiceRegistrationStatusResponse', (_message.Message,), {
  'DESCRIPTOR' : _SERVICEREGISTRATIONSTATUSRESPONSE,
  '__module__' : 'bootstrap_pb2'
  # @@protoc_insertion_point(class_scope:ServiceRegistrationStatusResponse)
  })
_sym_db.RegisterMessage(ServiceRegistrationStatusResponse)



_BOOTSTRAPSERVER = _descriptor.ServiceDescriptor(
  name='BootstrapServer',
  full_name='BootstrapServer',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=483,
  serialized_end=687,
  methods=[
  _descriptor.MethodDescriptor(
    name='RegisterService',
    full_name='BootstrapServer.RegisterService',
    index=0,
    containing_service=None,
    input_type=_SERVICEREGISTERREQUEST,
    output_type=_SERVICEREGISTRATIONSTATUSRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetServiceRegistrationStatus',
    full_name='BootstrapServer.GetServiceRegistrationStatus',
    index=1,
    containing_service=None,
    input_type=_SERVICEREGISTRATIONSTATUSREQUEST,
    output_type=_SERVICEREGISTRATIONSTATUSRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_BOOTSTRAPSERVER)

DESCRIPTOR.services_by_name['BootstrapServer'] = _BOOTSTRAPSERVER

# @@protoc_insertion_point(module_scope)



from hedra.core.graphs.stages import (
    Execute,
)

from hedra.core.graphs.hooks import (
    action
)


class ExecutePlaywrightStage(Execute):

    @action()
    async def open_page(self):
        return await self.client.playwright.goto('https://<url_here>')

    @action()
    async def click_item(self):
        return await self.client.playwright.click('[<selector_here>]')


import functools
from hedra.plugins.types.common.plugin_hook import PluginHook
from hedra.plugins.types.common.types import PluginHooks
from hedra.plugins.types.common.registrar import plugin_registrar


@plugin_registrar(PluginHooks.ON_PROCESS_EVENTS)
def process_events():

    def wrapper(func) -> PluginHook:

        @functools.wraps(func)
        def decorator(*args, **kwargs):

            return func(*args, **kwargs)
                
        return decorator

    return wrapper


from .leader_pb2 import (
        WorkerRegisterRequest,
        WorkerRegisterResponse,
        WorkerUpdateRequest,
        WorkerUpdateResponse,
        WorkerDeregisterRequest,
        WorkerDeregisterResponse,
        LeaderRegisterRequest,
        LeaderRegisterResponse,
        PollLeaderRequest,
        PollLeaderResponse,
        PipelineStageRequest,
        PipelineStageResponse,
        NewJobLeaderRequest,
        NewJobLeaderResponse,
        LeaderHeartbeatRequest,
        LeaderHeartbeatResponse,
        LeaderJobRequest,
        LeaderJobResponse
)

from .leader_pb2_grpc import (
    DistributedServerServicer,
    DistributedServerStub,
    add_DistributedServerServicer_to_server
)


from .transitions import (
    execute_to_setup_transition,
    execute_to_execute_transition,
    execute_to_optimize_transition,
    execute_to_teardown_transition,
    execute_to_analyze_transition,
    execute_to_checkpoint_transition,
    execute_to_wait_transition
)


import grpc
import os
import psutil
from .service import BootstrapService
from .proto import add_BootstrapServerServicer_to_server


class BootstrapServer:

    def __init__(self, config, leader_id) -> None:
        self.leader_id = leader_id

        self.session_logger.info('Initializing bootstrap server...')
        
        self._worker_threads = config.distributed_config.get(
            'max_workers',
            os.getenv(
                'MAX_THREAD_WORKERS',
                psutil.cpu_count(logical=False)
            )
        )

        self.server = grpc.aio.server()
        self.service = BootstrapService(config)

        add_BootstrapServerServicer_to_server(
            self.service,
            self.server
        )

        server_address = f'[::]:{self.service.bootstrap_port}'
        self.server.add_insecure_port(server_address)

        self.run_server = False

    async def run(self):
        self.session_logger.info(f'Bootstrap server serving on port - {self.service.bootstrap_port}')

        await self.server.start()
        self.service.running = True
        self.run_server = True

    async def stop_server(self):
        await self.service.stop(0)


import asyncio
from typing import Dict, Iterator, List, Union
from hedra.core.engines.client.config import Config
from hedra.core.engines.types.http import (
    MercuryHTTPClient,
    HTTPAction,
    HTTPResult
)
from hedra.core.engines.types.common.types import RequestTypes
from hedra.core.engines.types.common import Timeouts
from hedra.core.engines.client.store import ActionsStore
from hedra.logging import HedraLogger
from .base_client import BaseClient


class HTTPClient(BaseClient[MercuryHTTPClient, HTTPAction, HTTPResult]):

    def __init__(self, config: Config) -> None:
        super().__init__()

        self.session = MercuryHTTPClient(
            concurrency=config.batch_size,
            timeouts=Timeouts(
                total_timeout=config.request_timeout
            ),
            reset_connections=config.reset_connections
        )
        self.request_type = RequestTypes.HTTP
        self.client_type = self.request_type.capitalize()

        self.next_name = None
        self.intercept = False
        self.waiter = None
        self.actions: ActionsStore = None
        self.registered = {}

        self.logger = HedraLogger()
        self.logger.initialize()

    def __getitem__(self, key: str):
        return self.session.registered.get(key)

    async def get(
        self,
        url: str, 
        headers: Dict[str, str] = {}, 
        user: str = None,
        tags: List[Dict[str, str]] = [],
        redirects: int = 10
    ):

        request = HTTPAction(
            self.next_name,
            url,
            method='GET',
            headers=headers,
            data=None,
            user=user,
            tags=tags,
            redirects=redirects             
        )

        return await self._execute_action(request)

    async def post(
        self,
        url: str, 
        headers: Dict[str, str] = {}, 
        data: Union[dict, str, bytes, Iterator] = None,
        user: str = None,
        tags: List[Dict[str, str]] = [],
        redirects: int = 10
    ):

        request = HTTPAction(
            self.next_name,
            url,
            method='POST',
            headers=headers,
            data=data,
            user=user,
            tags=tags,
            redirects=redirects           
        )

        return await self._execute_action(request)

    async def put(
        self,
        url: str, 
        headers: Dict[str, str] = {}, 
        data: Union[dict, str, bytes, Iterator] = None,
        user: str = None,
        tags: List[Dict[str, str]] = [],
        redirects: int = 10
    ):

        request = HTTPAction(
            self.next_name,
            url,
            method='PUT',
            headers=headers,
            data=data,
            user=user,
            tags=tags,
            redirects=redirects
        )

        return await self._execute_action(request)

    async def patch(
        self,
        url: str, 
        headers: Dict[str, str] = {}, 
        data: Union[dict, str, bytes, Iterator] = None,
        user: str = None,
        tags: List[Dict[str, str]] = [],
        redirects: int = 10
    ):

        request = HTTPAction(
            self.next_name,
            url,
            method='PATCH',
            headers=headers,
            data=data,
            user=user,
            tags=tags,
            redirects=redirects
        )

        return await self._execute_action(request)

    async def delete(
        self, 
        url: str, 
        headers: Dict[str, str] = {}, 
        user: str = None,
        tags: List[Dict[str, str]] = [],
        redirects: int = 10
    ):
        request = HTTPAction(
            self.next_name,
            url,
            method='DELETE',
            headers=headers,
            data=None,
            user=user,
            tags=tags,
            redirects=redirects
        )

        return await self._execute_action(request)



from typing import Type, Callable, Awaitable, Any
from hedra.core.graphs.simple_context import SimpleContext
from hedra.core.graphs.hooks.hook_types.hook_type import HookType
from hedra.tools.filesystem import open
from .hook import Hook


class SaveHook(Hook):

    def __init__(
        self, 
        name: str, 
        shortname: str, 
        call: Callable[..., Awaitable[Any]], 
        key: str=None,
        checkpoint_filepath: str=None
    ) -> None:
        super().__init__(
            name, 
            shortname, 
            call, 
            hook_type=HookType.SAVE
        )

        self.context_key = key
        self.save_path = checkpoint_filepath


    async def call(self, context: SimpleContext) -> None:

        save_file = await open(self.save_path, 'w')
        await save_file.write(
            await self._call(
                context.get(self.context_key)
            )
        )

        await save_file.close()


        if context.get(self.context_key):
            context[self.context_key] = None


        

